Document: kdl
Title: Debian kdl Manual
Author: <insert document author here>
Abstract: This manual describes what kdl is
 and how it can be used to
 manage online manuals on Debian systems.
Section: unknown

Format: debiandoc-sgml
Files: /usr/share/doc/kdl/kdl.sgml.gz

Format: postscript
Files: /usr/share/doc/kdl/kdl.ps.gz

Format: text
Files: /usr/share/doc/kdl/kdl.text.gz

Format: HTML
Index: /usr/share/doc/kdl/html/index.html
Files: /usr/share/doc/kdl/html/*.html

  
